#include <iostream>
#include <cmath>

#include "datapoint.h"
#include "functions.h"

double   DataPoint::dp_s_Lscale( 10 );
double   DataPoint::dp_s_dTair( 0.1 );
double   DataPoint::dp_s_drelhum( 1 );

DataPoint::DataPoint( void ) :
  dp_status( ' ' ),
  dp_id(), dp_date(),
  dp_x( -1 ), dp_y( -1 ),
  dp_Ts( -abs_zero_offset ), dp_dTs( 0 ), dp_Tbg( -abs_zero_offset ), dp_dTbg( 0 ), dp_Tg( -abs_zero_offset ), dp_dTg( 0 ),
  dp_grad( 0 ), dp_dgrad( 0 ),
  dp_hconv( 0, 0 ), dp_hconv_b( 0, 0 )
{}

void DataPoint::read( std::istream &is )
{
  double    Tair, Ts, Tbg, Tg, grad, dgrad;
  
  is >> this->dp_id >> this->dp_date
     >> this->dp_x >> this->dp_y
     >> Tair >> this->dp_relhum
     >> Ts >> this->dp_dTs >> Tbg >> this->dp_dTbg
     >> Tg >> this->dp_dTg >> grad >> dgrad
     >> this->dp_status;

  this->dp_Tair = to_kelvin( Tair );
  this->dp_Ts = to_kelvin( Ts );
  this->dp_Tbg = to_kelvin( Tbg );
  this->dp_Tg = to_kelvin( Tg );
  this->dp_grad = 100 * grad; // Convert to C/cm to C/m
  this->dp_dgrad = 100 * dgrad; // Convert to C/cm to C/m

  this->dp_hconv = convective_heat_coefficient( this->dp_Ts, this->dp_dTs, this->dp_Tair, dp_s_dTair, dp_s_Lscale );
  this->dp_hconv_b = convective_heat_coefficient( this->dp_Tbg, this->dp_dTbg, this->dp_Tair, dp_s_dTair, dp_s_Lscale );

  return;
}

void DataPoint::print( std::ostream &os ) const
{
  os << "Point " << this->dp_id << ' ' << this->dp_status << " (" << this->dp_date << ") => [" << this->dp_x << ", " << this->dp_y << ']';

  return;
}

ValueError DataPoint::convectiveHeat( const ValueError &temp, const ValueError &hconv, const char *sep, int width ) const
{
  double     qconv, d_qconv;

  qconv = hconv.value * (temp.value - this->dp_Tair);
  d_qconv = std::sqrt( square((temp.value - this->dp_Tair) * hconv.error) + square(hconv.value) * (square(temp.error) + square(dp_s_dTair)) );

  return ( width > 0 ? ValueError(qconv, d_qconv, width) : ValueError(qconv, d_qconv, sep) );
}

ValueError DataPoint::ConvectiveHeat( const char *sep ) const
{
  return this->convectiveHeat( ValueError(this->dp_Ts, this->dp_dTs), this->dp_hconv, sep );
}

ValueError DataPoint::ConvectiveHeat( int width ) const
{
  return this->convectiveHeat( ValueError(this->dp_Ts, this->dp_dTs), this->dp_hconv, NULL, width );
}

ValueError DataPoint::evaporativeHeat( const ValueError &temp, const ValueError &hconv, const char *sep, int width ) const
{
  double       qeva, d_qeva, hvap, d_hvap, h_frac;
  ValueError   tmean( mean_temperature(temp.value, temp.error, this->dp_Tair, dp_s_dTair) );
  ValueError   psat_soil( saturation_pressure_over_liquid(temp.value, temp.error) ); // Saturation pressure at soil
  ValueError   psat_air(  saturation_pressure_over_liquid(this->dp_Tair, dp_s_dTair) ); // Saturation pressure in the air
  ValueError   pwat_air( partial_pressure(psat_air, ValueError(this->dp_relhum, dp_s_drelhum)) );   // Partial water pressure in the air
  ValueError   rhowat_soil( vapour_density(ValueError(temp.value, temp.error), psat_soil) );  // Water vapour density at soil (pwat_soil = psat_soil as we consider hum = 100% there
  ValueError   rhowat_air(  vapour_density(ValueError(this->dp_Tair, dp_s_dTair), pwat_air) );   // Water vapour density in the air

  if( rhowat_soil.value >= rhowat_air.value ) {
    ValueError   Kdiff_air( air_thermal_diffusivity(tmean.value, tmean.error, standard_Pair) );
    ValueError   Kdiff_vap( vapour_diffusion_coefficient(tmean.value, tmean.error, standard_Pair) );
    ValueError   lewis( Kdiff_air.value / Kdiff_vap.value, std::sqrt(square(Kdiff_air.error) + square(Kdiff_air.value * Kdiff_vap.error / Kdiff_vap.value)) / std::abs(Kdiff_vap.value) );
    ValueError   rho_air( air_density(tmean.value, standard_Pair), air_density_error(tmean.value, tmean.error, standard_Pair) );
    ValueError   cp_air( air_sh(tmean.value), air_sh_error(tmean.value, tmean.error) );

    h_frac = rho_air.value * cp_air.value * std::pow(lewis.value, 2./3.);
    hvap = hconv.value / h_frac;
    d_hvap = std::sqrt( square(hconv.error) + square(hconv.value) * (square(rho_air.error/rho_air.value) + square(cp_air.error/cp_air.value) +
								     4. * square(lewis.error/lewis.value) / 9.) ) / h_frac;

    qeva = wat_lh * hvap * (rhowat_soil.value - rhowat_air.value);
    d_qeva = wat_lh * std::sqrt( square(d_hvap * (rhowat_soil.value - rhowat_air.value)) + square(hvap) * (square(rhowat_soil.error) + square(rhowat_air.error)) );
  }
  else
    qeva = d_qeva = 0;
  
  return ( width > 0 ? ValueError(qeva, d_qeva, width) : ValueError(qeva, d_qeva, sep) );
}

ValueError DataPoint::EvaporativeHeat( const char *sep ) const
{
  return this->evaporativeHeat( ValueError(this->dp_Ts, this->dp_dTs), this->dp_hconv, sep );
}

ValueError DataPoint::EvaporativeHeat( int width ) const
{
  return this->evaporativeHeat( ValueError(this->dp_Ts, this->dp_dTs), this->dp_hconv, NULL, width );
}

ValueError DataPoint::radiativeHeat( const char *sep, int width ) const
{
  double          Tsky4, dTsky4, qrad, d_qrad, frac( relative_emissivity * sb_constant );
  ValueError      temp_bg( this->dp_Tbg, this->dp_dTbg );
  ValueError      heat_conv_b( this->convectiveHeat(temp_bg, this->dp_hconv_b, " ") );
  ValueError      heat_evap_b( this->evaporativeHeat(temp_bg, this->dp_hconv_b, " ") );

  Tsky4 = doublesquare(this->dp_Tbg) - (heat_conv_b.value + heat_evap_b.value) / frac;
  dTsky4 = std::sqrt( (4 * cube(this->dp_Tbg) * square(this->dp_dTbg)) + ((square(heat_conv_b.error) + square(heat_evap_b.error)) / square(frac)) );

  qrad = frac * (doublesquare(this->dp_Ts) - Tsky4);
  d_qrad = frac * std::sqrt( (4 * cube(this->dp_Ts) * square(this->dp_dTs)) + square(dTsky4) );

  return ( width > 0 ? ValueError(qrad, d_qrad, width) : ValueError(qrad, d_qrad, sep) );
}

ValueError DataPoint::RadiativeHeat( const char *sep ) const
{
  return this->radiativeHeat( sep );
}

ValueError DataPoint::RadiativeHeat( int width ) const
{
  return this->radiativeHeat( NULL, width );
}
