#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <list>

#include <TApplication.h>
#include <TGraphErrors.h>
#include <TMultiGraph.h>
#include <TCanvas.h>
#include <TStyle.h>
#include <TFitResult.h>
#include <TFitResultPtr.h>
#include <TH1.h>
#include <TLegend.h>
#include <TAxis.h>
#include <TVirtualFitter.h>

#include "fitter.h"
#include "heatpoint.h"
#include "residual.h"
#include "valueerror.h"
#include "functions.h"

namespace {

typedef std::pair<TH1D *, TH1D *>      HistoPair;

inline void save_images( TCanvas *canvas, const std::string &prefix )
{
  size_t          i, number;
  std::string     filename;
  const char     *extensions[] = { ".eps", ".png", ".pdf", ".ps" };

  number = sizeof(extensions) / sizeof(char *);
  for( i = 0; i < number; ++i ) {
    filename = prefix + extensions[i];
    canvas->SaveAs( filename.c_str() );
  }

  return;
}

Double_t linear_function( Double_t *x, Double_t *par )
{
  return par[0] + par[1] * x[0];
}

HistoPair calculate_residuals( const std::vector<HeatPoint> &points, const ValueError &a, const ValueError &b, int nbins = 100, const std::string what = "residual", const std::string title = "Residuals" )
{
  TH1D                                    *histo, *histo_w;
  double                                   rmin, rmax, min( 0 ), max( 0 );
  std::string                              name( what + "_histo" );
  std::vector<Residual>                    residuals;
  std::vector<Residual>::iterator          rIt;
  std::vector<HeatPoint>::const_iterator   pIt;

  residuals.reserve( points.size() );

  for( pIt = points.begin(); pIt != points.end(); ++pIt ) {
    Residual   r( Residual::calculateResidual(pIt->getTotalHeatValue(), pIt->getGradientValue(), a, b) );

    residuals.push_back( r );

    if( (min == 0) && (max == 0) ) {
      min = r.getMinimum();
      max = r.getMaximum();
    }
    else {
      rmin = r.getMinimum();
      rmax = r.getMaximum();

      if( min > rmin )
	min = rmin;

      if( max < rmax )
	max = rmax;
    }
  }

  histo = new TH1D( name.c_str(), title.c_str(), nbins, min, max );
  histo_w = new TH1D( (name + "_w").c_str(), (title + " (Weighted)").c_str(), nbins, min, max );

  for( rIt = residuals.begin(); rIt != residuals.end(); ++rIt ) {
    histo->Fill( rIt->getValue() );
    histo_w->Fill( rIt->getValue(), 1 / square(rIt->getError()) );
  }

  return HistoPair( histo, histo_w );
}

HistoPair calculate_ratios( const std::vector<HeatPoint> &points, const ValueError &a, const ValueError &b, int nbins = 100, const std::string &what = "ratios", const std::string &title = "Ratios" )
{
  TH1D                                   *histo, *histo_w;
  double                                  rmin, rmax, min( 0 ), max( 0 );
  std::string                             name( what + "_histo" );
  std::vector<Ratio>                      ratios;
  std::vector<HeatPoint>::const_iterator  pIt;
  std::vector<Ratio>::iterator            rIt;

  ratios.reserve( points.size() );

  for( pIt = points.begin(); pIt != points.end(); ++pIt ) {
    Ratio   r( Ratio::calculateRatio(pIt->getTotalHeatValue(), pIt->getGradientValue(), a) );

    ratios.push_back( r );

    if( (min == 0) && (max == 0) ) {
      min = r.getMinimum();
      max = r.getMaximum();
    }
    else {
      rmin = r.getMinimum();
      rmax = r.getMaximum();

      if( min > rmin )
	min = rmin;

      if( max < rmax )
	max = rmax;
    }
  }

  histo = new TH1D( name.c_str(), title.c_str(), nbins, min, max );
  histo_w = new TH1D( (name + "_w").c_str(), (title + " (Weighted)").c_str(), nbins, min, max );

  for( rIt = ratios.begin(); rIt != ratios.end(); ++rIt ) {
    histo->Fill( rIt->getValue() );
    histo_w->Fill( rIt->getValue(), (1 / square(rIt->getError())) );
  }

  return HistoPair( histo, histo_w );
}

std::string create_label( TFitResultPtr results )
{
  std::ostringstream     ostr;

  ostr << "k_{soil} = " << results->Value( 1 ) << " #pm " << std::setprecision( 2 ) << results->ParError( 1 );

  return ostr.str();
}

TH1D *totalheat_relative_error( const std::vector<HeatPoint> &points, const std::string &name = "relative", const std::string &title = "Relative error", int bins = 100 )
{
  size_t                  n, number( static_cast<std::size_t>(points.size()) );
  double                  rel, min, max;
  TH1D                   *histo;
  std::vector<double>     rd( points.size(), 0 );

  for( n = 0; n < number; ++n ) {
    rel = points[n].getTotalHeatRelativeError();
    rd[n] = rel;

    if( n == 0 )
      min = max = rel;
    else {
      if( min > rel )
	min = rel;

      if( max < rel )
	max = rel;
    }
  }

  histo = new TH1D( name.c_str(), title.c_str(), bins, min, max );

  for( n = 0; n < number; ++n )
    histo->Fill( rd[n] );

  return histo;
}

void separate_families( const std::vector<HeatPoint> &points, const ValueError &a, const ValueError &b )
{
  int                                      count_low( 0 ), count_middle1( 0 ), count_middle2( 0 ), count_high( 0 ), count_veryhigh( 0 );
  std::vector<HeatPoint>::const_iterator   pIt;
  std::string                              low_label, middle1_label, middle2_label, high_label, veryhigh_label;
  TFitResultPtr                            results;
  TGraphErrors                            *low( new TGraphErrors() ), *middle1( new TGraphErrors() ), *middle2( new TGraphErrors() ), *high( new TGraphErrors() ), *veryhigh( new TGraphErrors() );
  TMultiGraph                             *mg( new TMultiGraph() );
  TLegend                                 *legend( new TLegend(0.098, 0.75, 0.3, 0.9) );
  std::ofstream                            f1( "family_1.dat", std::ios::out ), f2( "family_2.dat", std::ios::out ), f3( "family_3.dat", std::ios::out );
  std::ofstream                            f4( "family_4.dat", std::ios::out ), f5( "family_5.dat", std::ios::out );

  for( pIt = points.begin(); pIt != points.end(); ++pIt ) {
    Ratio    r( Ratio::calculateRatio(pIt->getTotalHeatValue(), pIt->getGradientValue(), a) );

    if( r.getValue() <= 0.55 ) {
      low->SetPoint( count_low, pIt->getGradient(), pIt->getTotalHeat() );
      low->SetPointError( count_low, pIt->getGradientError(), pIt->getTotalHeatError() );

      pIt->printCoord( f1 ) << std::endl;

      count_low += 1;
    }
    else if( r.getValue() <= 1.05 ) {
      middle1->SetPoint( count_middle1, pIt->getGradient(), pIt->getTotalHeat() );
      middle1->SetPointError( count_middle1, pIt->getGradientError(), pIt->getTotalHeatError() );

      pIt->printCoord( f2 ) << std::endl;

      count_middle1 += 1;
    }
    else if( r.getValue() <= 1.55 ) {
      middle2->SetPoint( count_middle2, pIt->getGradient(), pIt->getTotalHeat() );
      middle2->SetPointError( count_middle2, pIt->getGradientError(), pIt->getTotalHeatError() );

      pIt->printCoord( f3 ) << std::endl;

      count_middle2 += 1;
    }
    else if( r.getValue() <= 2.55 ) {
      high->SetPoint( count_high, pIt->getGradient(), pIt->getTotalHeat() );
      high->SetPointError( count_high, pIt->getGradientError(), pIt->getTotalHeatError() );

      pIt->printCoord( f4 ) << std::endl;

      count_high += 1;
    }
    else {
      veryhigh->SetPoint( count_veryhigh, pIt->getGradient(), pIt->getTotalHeat() );
      veryhigh->SetPointError( count_veryhigh, pIt->getGradientError(), pIt->getTotalHeatError() );

      pIt->printCoord( f5 ) << std::endl;

      count_veryhigh += 1;
    }
  }

  f1.close();
  f2.close();
  f3.close();
  f4.close();
  f5.close();

  low->SetMarkerStyle( kFullTriangleDown );
  low->SetMarkerColor( kBlue );
  low->SetMarkerSize( 1.2 );
  low->SetLineColor( kBlue );

  middle1->SetMarkerStyle( kFullSquare );
  middle1->SetMarkerColor( kGreen );
  middle1->SetMarkerSize( 1.2 );
  middle1->SetLineColor( kGreen );

  middle2->SetMarkerStyle( kFullSquare );
  middle2->SetMarkerColor( kYellow );
  middle2->SetMarkerSize( 1.2 );
  middle2->SetLineColor( kYellow );

  high->SetMarkerStyle( kFullTriangleUp );
  high->SetMarkerColor( kOrange );
  high->SetMarkerSize( 1.2 );
  high->SetLineColor( kOrange );

  veryhigh->SetMarkerStyle( kFullTriangleUp );
  veryhigh->SetMarkerColor( kRed );
  veryhigh->SetMarkerSize( 1.2 );
  veryhigh->SetLineColor( kRed );

  results = low->Fit( "pol1", "S" );
  low_label = create_label( results );
  
  results = middle1->Fit( "pol1", "S" );
  middle1_label = create_label( results );
  
  results = middle2->Fit( "pol1", "S" );
  middle2_label = create_label( results );
  
  results = high->Fit( "pol1", "S" );
  high_label = create_label( results );
  
  results = veryhigh->Fit( "pol1", "S" );
  veryhigh_label = create_label( results );

  low->GetFunction( "pol1" )->SetLineColor( kBlue );
  middle1->GetFunction( "pol1" )->SetLineColor( kGreen );
  middle2->GetFunction( "pol1" )->SetLineColor( kYellow );
  high->GetFunction( "pol1" )->SetLineColor( kOrange );
  veryhigh->GetFunction( "pol1" )->SetLineColor( kRed );

  legend->AddEntry( low, low_label.c_str(), "p" );
  legend->AddEntry( middle1, middle1_label.c_str(), "p" );
  legend->AddEntry( middle2, middle2_label.c_str(), "p" );
  legend->AddEntry( high, high_label.c_str(), "p" );
  legend->AddEntry( veryhigh, veryhigh_label.c_str(), "p" );

  mg->Add( low );
  mg->Add( middle1 );
  mg->Add( middle2 );
  mg->Add( high );
  mg->Add( veryhigh );

  TCanvas         *families( new TCanvas("families", "Families", 0, 0, 1000, 600) );

  families->SetGrid( true );
  mg->Draw( "AP" );
  mg->GetXaxis()->SetTitle( "#nablaT_{s} (K/m)" );
  mg->GetYaxis()->SetTitle( "(q_{conv} + q_{rad} + q_{evap}) (W/m^{2})" );
  mg->Draw( "AP" );
  legend->Draw();

  save_images( families, "images/families" );

  return;
}

} // anonymous namespace

int execute_fits( const std::vector<HeatPoint> &points, bool show )
{
  int                                      result( 0 );
  int                                      i, count_all, count_deep, count_shallow, count_night, interval_number( 500 );
  double                                   x, delta;
  std::ofstream                            ofs( "fit_results.txt", std::ios::out );
  ValueError                               as( 0, 0, " +/- " ), bs( 0, 0, " +/- " ), an( 0, 0, " +/- " ), bn( 0, 0, " +/- " );
  ValueError                               ane( 0 , 0, " +/- " ), bne( 0, 0, " +/- " ), ase( 0, 0, " +/- " ), bse( 0, 0, " +/- " );
  TApplication                             theApp( "HeatTransfer", 0, NULL );
  TFitResultPtr                            results;
  TGraphErrors                            *all_points( new TGraphErrors() ), *deep_points( new TGraphErrors() ), *shallow_points( new TGraphErrors() ), *night_points( new TGraphErrors() );
  TGraphErrors                            *extrapolated_shallow( new TGraphErrors() ), *extrapolated_night( new TGraphErrors() );
  TMultiGraph                             *mg( new TMultiGraph() ), *shallow_mg( new TMultiGraph() ), *night_mg( new TMultiGraph() );
  TMultiGraph                             *extrapolated_shallow_mg( new TMultiGraph() ), *extrapolated_night_mg( new TMultiGraph() );
  TF1                                     *fitfun;
  const char                               xlabel[] = "#nablaT_{s} (K/m)", ylabel[] = "(q_{conv} + q_{rad} + q_{evap}) (W/m^{2})";
  std::vector<HeatPoint>                   deep_vector, shallow_vector, night_vector;
  std::vector<HeatPoint>::const_iterator   pIt;

  count_all = count_deep = count_shallow = count_night = 0;
  deep_vector.reserve( points.size() );
  shallow_vector.reserve( points.size() );
  night_vector.reserve( points.size() );

  for( pIt = points.begin(); pIt != points.end(); ++pIt ) {
    if( pIt->isGood() ) {
      all_points->SetPoint( count_all, pIt->getGradient(), pIt->getTotalHeat() );
      all_points->SetPointError( count_all, pIt->getGradientError(), pIt->getTotalHeatError() );
      count_all += 1;

      if( pIt->isDeep() ) {
	deep_points->SetPoint( count_deep, pIt->getGradient(), pIt->getTotalHeat() );
	deep_points->SetPointError( count_deep, pIt->getGradientError(), pIt->getTotalHeatError() );
	count_deep += 1;

	deep_vector.push_back( *pIt );
      }
      else if( pIt->isShallow() ) {
	shallow_points->SetPoint( count_shallow, pIt->getGradient(), pIt->getTotalHeat() );
	shallow_points->SetPointError( count_shallow, pIt->getGradientError(), pIt->getTotalHeatError() );

	extrapolated_shallow->SetPoint( count_shallow, pIt->getSurfaceTemperature(), pIt->getGradientIntercept() );
	extrapolated_shallow->SetPointError( count_shallow, pIt->getSurfaceTemperatureError(), pIt->getGradientInterceptError() );
	
	count_shallow += 1;

	shallow_vector.push_back( *pIt );

	if( pIt->isNightPoint() ) {
	  night_points->SetPoint( count_night, pIt->getGradient(), pIt->getTotalHeat() );
	  night_points->SetPointError( count_night, pIt->getGradientError(), pIt->getTotalHeatError() );

	  extrapolated_night->SetPoint( count_night, pIt->getSurfaceTemperature(), pIt->getGradientIntercept() );
	  extrapolated_night->SetPointError( count_night, pIt->getSurfaceTemperatureError(), pIt->getGradientInterceptError() );
	  
	  count_night += 1;

	  night_vector.push_back( *pIt );
	  //	  std::cout << "Night point: " << pIt->getId() << " -> " << pIt->getGradient() << " +/- " << pIt->getGradientError() << std::endl;
	}
      }
    }
  }

  ofs << "All points: " << count_all << std::endl
      << "Deep points: " << count_deep << std::endl
      << "Shallow points: " << count_shallow << std::endl
      << "Night points: " << count_night << std::endl;

  all_points->SetMarkerStyle( kFullCircle );
  all_points->SetMarkerSize( 1.2 );
  all_points->SetMarkerColor( kBlack );
  all_points->SetLineColor( kBlack );

  deep_points->SetMarkerStyle( kFullTriangleDown );
  deep_points->SetMarkerSize( 1.2 );
  deep_points->SetMarkerColor( kBlack );
  deep_points->SetLineColor( kBlack );

  shallow_points->SetMarkerStyle( kFullTriangleUp );
  shallow_points->SetMarkerSize( 1.2 );
  shallow_points->SetMarkerColor( kRed );
  shallow_points->SetLineColor( kRed );

  night_points->SetMarkerStyle( kFullTriangleUp );
  night_points->SetMarkerSize( 1.2 );
  night_points->SetMarkerColor( kBlue );
  night_points->SetLineColor( kBlue );

  extrapolated_shallow->SetMarkerStyle( kFullTriangleUp );
  extrapolated_shallow->SetMarkerSize( 1.2 );
  extrapolated_shallow->SetMarkerColor( kRed );
  extrapolated_shallow->SetLineColor( kRed );

  extrapolated_night->SetMarkerStyle( kFullTriangleUp );
  extrapolated_night->SetMarkerSize( 1.2 );
  extrapolated_night->SetMarkerColor( kBlue );
  extrapolated_night->SetLineColor( kBlue );

  results = shallow_points->Fit( "pol1", "SWF" );
  fitfun = shallow_points->GetFunction( "pol1" );
  fitfun->SetLineColor( kRed );

  as.value = results->Value( 0 ); as.error = results->ParError( 0 );
  bs.value = results->Value( 1 ); bs.error = results->ParError( 1 );

  ofs << "Shallow point results:" << std::endl << "\ta = " << as << std::endl << "\tb = " << bs << std::endl
      << "\tCorrelation factor: " << shallow_points->GetCorrelationFactor()
      << " over " << count_shallow << " points." << std::endl;

  TGraphErrors    *shallow_interval( new TGraphErrors(interval_number) );

  delta = (fitfun->GetXmax() - fitfun->GetXmin()) / interval_number;
  for( i = 0, x = fitfun->GetXmin(); i < interval_number; ++i, x += delta )
    shallow_interval->SetPoint( i, x, 0 );

  TVirtualFitter::GetFitter()->GetConfidenceIntervals( shallow_interval );

  shallow_interval->SetLineColorAlpha( kRed - 10, 0.35 );
  shallow_interval->SetLineWidth( 2 );

  shallow_mg->Add( shallow_interval, "P" );
  shallow_mg->Add( shallow_points, "P" );

  HistoPair shallow_residuals( calculate_residuals(shallow_vector, as, bs, 100, "shallow", "Shallow points residuals") );
  HistoPair shallow_ratios( calculate_ratios(shallow_vector, as, bs, 250) );

  shallow_residuals.first->Fit( "gaus" );
  shallow_residuals.second->Fit( "gaus" );

  results = night_points->Fit( "pol1", "SWF" );
  fitfun = night_points->GetFunction( "pol1" );
  fitfun->SetLineColor( kBlue );

  an.value = results->Value( 0 ); an.error = results->ParError( 0 );
  bn.value = results->Value( 1 ); bn.error = results->ParError( 1 );

  ofs << "Night points results: " << std::endl << "\ta = " << an << std::endl << "\tb = " << bn << std::endl
      << "\tCorrelation factor: " << night_points->GetCorrelationFactor()
      << " over " << count_night << " points." << std::endl;

  TGraphErrors    *night_interval( new TGraphErrors(interval_number) );

  delta = (fitfun->GetXmax() - fitfun->GetXmin()) / interval_number;
  for( i = 0, x = fitfun->GetXmin(); i < interval_number; ++i, x += delta )
    night_interval->SetPoint( i, x, 0 );

  TVirtualFitter::GetFitter()->GetConfidenceIntervals( night_interval );

  night_interval->SetLineColorAlpha( kBlue - 10, 0.35 );
  night_interval->SetLineWidth( 2 );

  night_mg->Add( night_interval, "P" );
  night_mg->Add( night_points, "P" );

  results = extrapolated_shallow->Fit( "pol1", "SFW" );
  fitfun = extrapolated_shallow->GetFunction( "pol1" );
  fitfun->SetLineColor( kRed );

  ase.value = results->Value( 0 ); ase.error = results->ParError( 0 );
  bse.value = results->Value( 1 ); bse.error = results->ParError( 1 );
  
  ofs << "Shallow extrapolated fit:" << std::endl << "\ta = " << ase << std::endl << "\tb = " << bse << std::endl
      << "\tCorrelation factor: " << extrapolated_shallow->GetCorrelationFactor()
      << " over " << count_shallow << " points." << std::endl;

  TGraphErrors    *extrapolated_shallow_interval( new TGraphErrors(interval_number) );

  delta = (fitfun->GetXmax() - fitfun->GetXmin()) / interval_number;
  for( i = 0, x = fitfun->GetXmin(); i < interval_number; ++i, x += delta )
    extrapolated_shallow_interval->SetPoint( i, x, 0 );

  TVirtualFitter::GetFitter()->GetConfidenceIntervals( extrapolated_shallow_interval );

  extrapolated_shallow_interval->SetLineColorAlpha( kRed - 10, 0.35 );
  extrapolated_shallow_interval->SetLineWidth( 2 );

  extrapolated_shallow_mg->Add( extrapolated_shallow_interval, "P" );
  extrapolated_shallow_mg->Add( extrapolated_shallow, "P" );

  results = extrapolated_night->Fit( "pol1", "SFW" );
  fitfun = extrapolated_night->GetFunction( "pol1" );
  fitfun->SetLineColor( kBlue );
  
  ane.value = results->Value( 0 ); ane.error = results->ParError( 0 );
  bne.value = results->Value( 1 ); bne.error = results->ParError( 1 );

  ofs << "Night extrapolated fit:" << std::endl << "\ta = " << ane << std::endl << "\tb = " << bne << std::endl
      << "\tCorrelation factor: " << extrapolated_night->GetCorrelationFactor()
      << " over " << count_night << " points." << std::endl;

  TGraphErrors    *extrapolated_night_interval( new TGraphErrors(interval_number) );

  delta = (fitfun->GetXmax() - fitfun->GetXmin()) / interval_number;
  for( i = 0, x = fitfun->GetXmin(); i < interval_number; ++i, x += delta )
    extrapolated_night_interval->SetPoint( i, x, 0 );

  TVirtualFitter::GetFitter()->GetConfidenceIntervals( extrapolated_night_interval );

  extrapolated_night_interval->SetLineColorAlpha( kBlue - 10, 0.35 );
  extrapolated_night_interval->SetLineWidth( 2 );

  extrapolated_night_mg->Add( extrapolated_night_interval, "P" );
  extrapolated_night_mg->Add( extrapolated_night, "P" );

  HistoPair night_residuals( calculate_residuals(night_vector, an, bn, 100, "night", "Night points residuals") );

  night_residuals.first->Fit( "gaus" );
  night_residuals.second->Fit( "gaus" );
  
  mg->Add( deep_points );
  mg->Add( shallow_points );
  mg->Add( night_points );

  //  gStyle->SetOptFit( 1110 );
  gStyle->SetOptStat( false );

  separate_families( shallow_vector, as, bs );

  TH1D *shallow_relerr = totalheat_relative_error( shallow_vector, "relsha", "Shallow relative errors" );
  TH1D *night_relerr = totalheat_relative_error( night_vector, "relnig", "Night relative errors" );

  TCanvas       *relative_canvas( new TCanvas("relative_canvas", "Relative errors", 800, 800, 800, 400) );

  relative_canvas->Divide( 2, 1 );

  relative_canvas->cd( 1 );
  relative_canvas->GetPad( 1 )->SetGrid( true );
  shallow_relerr->Draw();

  relative_canvas->cd( 2 );
  relative_canvas->GetPad( 2 )->SetGrid( true );
  night_relerr->Draw();

  save_images( relative_canvas, "images/relative_errors" );

  TCanvas       *rubbish( new TCanvas("rubbish", "Heat", 0, 0, 1024, 512) );

  rubbish->Divide( 2, 1 );

  rubbish->cd( 1 );
  rubbish->GetPad( 1 )->SetGrid( true );
  all_points->Draw( "AP" );
  all_points->GetXaxis()->SetTitle( xlabel );
  all_points->GetYaxis()->SetTitle( ylabel );

  rubbish->cd( 2 );
  rubbish->GetPad( 2 )->SetGrid( true );
  deep_points->Draw( "AP" );
  deep_points->GetXaxis()->SetTitle( xlabel );
  deep_points->GetYaxis()->SetTitle( ylabel );

  save_images( rubbish, "images/points" );

  TCanvas       *shallow_canvas( new TCanvas("shallow_canvas", "Shallow points", 0, 0, 1000, 600) );

  shallow_canvas->SetGrid( true );
  shallow_mg->Draw( "AP" );
  shallow_mg->GetXaxis()->SetTitle( xlabel );
  shallow_mg->GetYaxis()->SetTitle( ylabel );
  shallow_mg->Draw( "AP" );

  save_images( shallow_canvas, "images/shallow_points" );

  TCanvas       *night_canvas( new TCanvas("night_canvas", "Night points", 0, 400, 1000, 600) );

  night_canvas->SetGrid( true );
  night_mg->Draw( "AP" );
  night_mg->GetXaxis()->SetTitle( xlabel );
  night_mg->GetYaxis()->SetTitle( ylabel );
  night_mg->Draw( "AP" );

  save_images( night_canvas, "images/night_points" );

  TCanvas      *extrapolated_shallow_canvas( new TCanvas("extrapolated_shallow_canvas", "Extrapolated shallow temperatures", 600, 400, 1000, 600) );

  extrapolated_shallow_canvas->SetGrid( true );
  extrapolated_shallow_mg->Draw( "AP" );
  extrapolated_shallow_mg->GetXaxis()->SetTitle( "Surface temperature (K)" ); //  (^{o}C)" );
  extrapolated_shallow_mg->GetYaxis()->SetTitle( "Extrapolated temperature (K)" ); // (^{o}C)" );
  extrapolated_shallow_mg->Draw( "AP" );

  save_images( extrapolated_shallow_canvas, "images/extrapolated_shallow" );

  TCanvas      *extrapolated_night_canvas( new TCanvas("extrapolated_canvas", "Extrapolated night temperatures", 600, 400, 1000, 600) );

  extrapolated_night_canvas->SetGrid( true );
  // extrapolated_night->SetMinimum( 0 );
  extrapolated_night_mg->Draw( "AP" );
  extrapolated_night_mg->GetXaxis()->SetTitle( "Surface temperature (K)" ); //  (^{o}C)" );
  extrapolated_night_mg->GetYaxis()->SetTitle( "Extrapolated temperature (K)" ); // (^{o}C)" );
  extrapolated_night_mg->Draw( "AP" );

  save_images( extrapolated_night_canvas, "images/extrapolated_night" );

  TCanvas       *cuollacuolla( new TCanvas("cuollacuolla", "Heat", 1024, 0, 1000, 600) );

  cuollacuolla->SetGrid( true );
  mg->Draw( "AP" );
  mg->GetXaxis()->SetTitle( xlabel );
  mg->GetYaxis()->SetTitle( ylabel );
  mg->Draw( "AP" );

  save_images( cuollacuolla, "images/all_points" );

  TCanvas       *shres_canvas( new TCanvas("shres_canvas", "Shallow Residuals", 1024, 600, 800, 600) );

  shres_canvas->Divide( 1, 2 );

  shres_canvas->cd( 1 );
  shres_canvas->GetPad( 1 )->SetGrid( true );
  shallow_residuals.first->Draw();

  shres_canvas->cd( 2 );
  shres_canvas->GetPad( 2 )->SetGrid( true );
  shallow_residuals.second->Draw();

  save_images( shres_canvas, "images/shallow_residuals" );

  TCanvas       *shrat_canvas( new TCanvas("shrat_canvas", "Shallow Ratios", 1024, 600, 800, 600) );

  shrat_canvas->Divide( 1, 2 );

  shrat_canvas->cd( 1 );
  shrat_canvas->GetPad( 1 )->SetGrid( true );
  shallow_ratios.first->Draw();

  shrat_canvas->cd( 2 );
  shrat_canvas->GetPad( 2 )->SetGrid( true );
  shallow_ratios.second->Draw();

  save_images( shrat_canvas, "images/shallow_ratios" );

  TCanvas       *nires_canvas( new TCanvas("nires_canvas", "Night Residuals", 1024, 600, 800, 600) );

  nires_canvas->Divide( 1, 2 );

  nires_canvas->cd( 1 );
  nires_canvas->GetPad( 1 )->SetGrid( true );
  night_residuals.first->Draw();

  nires_canvas->cd( 2 );
  nires_canvas->GetPad( 2 )->SetGrid( true );
  night_residuals.second->Draw();

  save_images( nires_canvas, "images/night_residuals" );

  if( show )
    theApp.Run();

  return result;
}
