#include <cmath>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <list>
#include <vector>

#include "datapoint.h"
#include "functions.h"

namespace {

typedef  std::vector<DataPoint>     PointVector;

void read_data( std::istream &is, std::vector<DataPoint> &points )
{
  std::list<DataPoint>  point_list;
  
  while( is.good() ) {
    DataPoint             dp;

    is >> dp;

    if( is.good() && dp.good() )
      point_list.push_back( dp );
  }

  if( !point_list.empty() )
    points = std::vector<DataPoint>( std::begin(point_list), std::end(point_list) );

  return;
}

int flusso_completo( std::istream &is, std::ostream &os )
{
  int                    result( 0 );
  PointVector            points;
  PointVector::iterator  pIt;

  if( is.good() ) {
    read_data( is, points );

    std::cerr << "Read <" << points.size() << "> points..." << std::endl;
    os << std::setw( 8 ) << std::left << "# Id" << std::setw( 2 ) << 'F'
       << std::setw( 21 ) << std::left << "Date"
       << std::setw( 9 ) << "X" << std::setw( 10 ) << "Y" << std::right
       << std::setw( 8 ) << "Ts" << std::setw( 9 ) << "d(Ts)"
       << std::setw( 8 ) << "Tg" << std::setw( 9 ) << "d(Tg)"
       << std::setw( 9 ) << "dT/dz" << std::setw( 9 ) << "d(dT/dz)"
       << std::setw( 10 ) << "Qc" << std::setw( 10 ) << "d(Qc)"
       << std::setw( 10 ) << "Qe" << std::setw( 10 ) << "d(Qe)"
       << std::setw( 10 ) << "Qr" << std::setw( 10 ) << "d(Qr)"
       << std::setw( 10 ) << "Qtot" << std::setw( 10 ) << "d(Qtot)" << std::endl;
    
    for( pIt = points.begin(); pIt != points.end(); ++pIt ) {
      if( pIt->good() ) {
	double       qtot, d_qtot;
	ValueError   Qconv( pIt->ConvectiveHeat(9) ), Qevap( pIt->EvaporativeHeat(9) ), Qrad( pIt->RadiativeHeat(9) );

	qtot = Qconv.value + Qevap.value + Qrad.value;
	d_qtot = std::sqrt( square(Qconv.error) + square(Qevap.error) + square(Qrad.error) );

	os << std::setfill( ' ' ) << std::setw( 7 ) << std::left << pIt->getId() << ' ' << std::setw( 2 ) << pIt->getFlag()
	   << pIt->getDate() << ' ' << std::right << std::fixed
	   << std::setprecision( 1 ) << std::setw( 9 ) << pIt->getX() << ' '
	   << std::setprecision( 1 ) << std::setw( 9 ) << pIt->getY() << ' '
	   << std::setprecision( 4 ) << std::right
	   << std::setw( 8 ) << pIt->getTs( 8 ) << ' '
	   << std::setw( 8 ) << pIt->getTg( 8 ) << ' '
	   << std::setw( 8 ) << pIt->getGrad( 8 ) << ' '
	   << std::setw( 9 ) << Qconv << ' ' << std::setw( 9 ) << Qevap << ' ' << std::setw( 9 ) << Qrad << ' '
	   << std::setw( 10 ) << qtot << std::setw( 9 ) << d_qtot << std::endl;
      }
    }
  }

  return result;

}

} // anonymous namespace

int flusso_completo( const char *input, const char *output )
{
  int                    result;
  std::ifstream          ifs( input, std::ios::in );
  std::ofstream          ofs;

  if( output == NULL )
    result = flusso_completo( ifs, std::cout );
  else {
    ofs.open( output, std::ios::out );

    result = flusso_completo( ifs, ofs );
  }

  return result;
}
