#include <iostream>
#include <cmath>

#include "constants.h"
#include "functions.h"
#include "datapoint.h"
#include "valueerror.h"

namespace {

const double     nu[3][2] = { { 0.54, 0.25 }, { 0.15, 1./3. }, { 0.27, 0.25 } };

inline int nusselt_index( double Ts, double Tair, double rayleigh )
{
  int         id = 2;

   if( Ts > Tair ) {
    if( rayleigh < 1e7 )
      id = 0;
    else
      id = 1;
  }

  return id;
}

} // anonymous namespace

ValueError air_thermal_diffusivity( double temp, double d_temp, double pres )
{
  double     Kconv_air( air_thermal_conductivity(temp) ), d_Kconv_air( air_thermal_conductivity_error(d_temp) );
  double     density_air( air_density(temp, pres) ), d_density_air( air_density_error(temp, d_temp, pres) );
  double     sh_air( air_sh(temp) ), d_sh_air( air_sh_error(temp, d_temp) );
  double     Kdiff_air( Kconv_air / (density_air * sh_air) );
  double     d_Kdiff_air;

  d_Kdiff_air = std::sqrt( square(d_Kconv_air / (density_air * sh_air)) +
			   square(Kconv_air * d_density_air / (sh_air * square(density_air))) +
			   square(Kconv_air * d_sh_air / (density_air * square(sh_air))) );

  return ValueError( Kdiff_air, d_Kdiff_air );
}

ValueError convective_heat_coefficient( double Ts, double dTs, double Tair, double dTair, double scale )
{
  int            id;
  double         ray_const, rayleigh, d_rayleigh, nusselt, d_nusselt, hconv, d_hconv;
  ValueError     tmean( mean_temperature(Ts, dTs, Tair, dTair) );
  double         Kconv_air( air_thermal_conductivity(tmean.value) ), d_Kconv_air( air_thermal_conductivity_error(tmean.error) );
  ValueError     Kdiff_air( air_thermal_diffusivity(tmean.value, tmean.error, standard_Pair) );

  ray_const = grav * beta_air * cube(scale) / (kvisc_air * Kdiff_air.value);
  rayleigh = ray_const * std::abs(Ts - Tair);
  d_rayleigh = ray_const * std::sqrt( square(dTs) + square(dTair) + square(((Ts - Tair) * Kdiff_air.error) / Kdiff_air.value) );

  id = nusselt_index( Ts, Tair, rayleigh );
  nusselt = nu[id][0] * std::pow( rayleigh, nu[id][1] );
  d_nusselt = d_rayleigh * std::abs(nu[id][0] * nu[id][1] * std::pow(rayleigh, nu[id][1] - 1));

  hconv = nusselt * Kconv_air / scale;
  d_hconv = std::sqrt( square(d_Kconv_air * nusselt) + square(d_nusselt * Kconv_air) ) / scale;

  return ValueError( hconv, d_hconv );
}

ValueError saturation_pressure_over_liquid( double temp, double d_temp )
{
  static const double   c1( 611.22 ), c2( 17.67 ), tamb( 29.65 );
  double                Tstar, psat, d_psat;

  Tstar = (temp - abs_zero_offset) / (temp - tamb);

  psat = c1 * std::exp( c2 * Tstar );
  d_psat = d_temp * std::abs( c2 * psat * (abs_zero_offset - tamb) / square(temp - tamb) );

  return ValueError( psat, d_psat );
}

ValueError vapour_diffusion_coefficient( double temp, double d_temp, double press )
{
  static const double   c1( 2.11e-5 ), c2( 1.94 );
  double                c3( c1 * standard_Pair / press ), trid( temp / abs_zero_offset );

  return ValueError( c3 * std::pow(trid, c2), d_temp * std::abs(c3 * c2 * std::pow(trid, c2 - 1) / abs_zero_offset) );
}
