#ifndef FUNCTIONS_H_LOADED
#define FUNCTIONS_H_LOADED

#include <cmath>
#include <iomanip>

#include "constants.h"
#include "valueerror.h"

#define  CC1      7.1128e-5

class DataPoint;

template <typename Type>
inline Type square( Type num ) { return num * num; }

template <typename Type>
inline Type cube( Type num ) { return num * num * num; }

template <typename Type>
inline Type doublesquare( Type num ) { return num * num * num * num; }

inline double to_kelvin( double temp )
{
  return temp + abs_zero_offset;
}

inline double to_celsius( double temp )
{
  return temp - abs_zero_offset;
}

inline double air_thermal_conductivity( double temp )
{
  return 0.023807 + CC1 * to_celsius(temp);  // Eq. 2.5 Jacobson, 2005
}

inline double air_thermal_conductivity_error( double d_temp )
{
  return CC1 * d_temp;
}

inline double vapour_thermal_conductivity( double temp )
{
  return 0.015606 + 8.3680e-5 * to_celsius(temp);  // Eq.2.6 Jacobson, 2005
}

inline double air_density( double temp, double pres )
{
  return pres / (air_R * temp);
}

inline double air_density_error( double temp, double d_temp, double pres )
{
  return d_temp * std::abs(pres / (air_R * temp * temp));
}

inline double air_sh( double temp ) // Specific heat of dry air at constant pressure (J/Kg K)
{
  return 1004.67; // At 298 K
}

inline double air_sh_error( double, double )
{
  return 0;
}

inline ValueError mean_temperature( double t1, double dt1, double t2, double dt2 )
{
  return ValueError( (t1 + t2) / 2, std::sqrt(square(dt1) + square(dt2)) / 2);
}

inline ValueError partial_pressure( const ValueError &psat, const ValueError &relhum )
{
  return ValueError( psat.value * relhum.value / 100, std::sqrt(square(psat.value * relhum.error) + square(psat.error * relhum.value)) / 100 );
}

inline ValueError vapour_density( const ValueError &temp, const ValueError &press )
{
  return ValueError( press.value / (wat_R * temp.value), std::sqrt(square(press.error) + square(press.value * temp.error / temp.value)) / (wat_R * temp.value) );
}

ValueError air_thermal_diffusivity( double temp, double d_temp, double pres );
ValueError convective_heat_coefficient( double ts, double dts, double ta, double dta, double scale );
ValueError saturation_pressure_over_liquid( double temp, double d_temp );
ValueError vapour_diffusion_coefficient( double temp, double d_temp, double press );

#undef CC1

#endif /* FUNCTIONS_H_LOADED */

// Local Variables:
// mode: c++
// End:
