program calcflussi
  ! Calcola i flussi di calore nei punti del file misure.dat
  !
  ! Author: Giovanni Macedonio
  !
  ! Version: 1.0
  !
  ! Date: 22-JUN-2018
  !
  use inpout
  use m_heatpar
  use m_airvapor
  implicit none
  integer :: narg
  integer :: nout
  integer(8), allocatable :: epoch(:)
  real(8), allocatable :: x(:),y(:)  ! Utm coordinates
  real(8), allocatable :: Ts(:),Ta(:),Tb(:),Toff(:),Tgrad(:),rhum(:)
  real(8), parameter :: Pair_standard=1.01325e5
  character(len=255) :: finp,fout
  type(RECLIST), target :: file
  class(RECLIST), pointer :: curr
  integer :: i,ndat,status
  real(8) :: q_soil,q_conv,q_conv_b,q_rad,q_evap,q_evap_b
  real(8) :: h_conv,h_conv_b,h_vapor,h_vapor_b
  real(8) :: lscale,tmean,tmean_b,pair,rhows,rhowa,rhowb,psats,psata,psatb,pwata
  real(8) :: lewis,lewis_b,err
  real(8) :: Tsky4
  logical :: correct_tsky_conv=.true.
  logical :: correct_tsky_evap=.true.

  ! Get the command line
  narg = iargc()
  if(narg < 2) then
     write(*,'(''Usage: calcflussi misure.dat flussi.out'')')
     call exit(1)
  end if
  !
  call getarg(1,finp)
  call getarg(2,fout)

  call file%load(finp,status)
  if(status /= 0) then
     write(*,'(''Cannot open file: '',a)') trim(finp)
     call exit(1)
  end if
  call file%remove_comments('#')
  ndat = file%nrec()
  ! call file%print

  ! Allocate memory
  allocate(x(ndat))
  allocate(y(ndat))
  allocate(Ts(ndat))
  allocate(Ta(ndat))
  allocate(Tb(ndat))
  allocate(Toff(ndat))
  allocate(Tgrad(ndat))
  allocate(rhum(ndat))
  allocate(epoch(ndat))

  ! Read
  ndat = 0
  curr => file
  do while(associated(curr))
     if(.not.associated(curr%rec)) exit
     ndat = ndat+1
     read(curr%rec,*) epoch(ndat),x(ndat),y(ndat),Ts(ndat),Ta(ndat), &
          Tb(ndat),Toff(ndat),Tgrad(ndat),rhum(ndat)
     curr => curr%next
  end do

  ! Free memory
  call file%free

  ! Convert temperatures in Kelvin
  Ta = kelvin(Ta)
  Ts = kelvin(Ts)
  Tb = kelvin(Tb)
  Toff = kelvin(Toff)

  ! @@@@@@@@@@@@@@@@@@@@@@
  !  Ts = Toff

  ! Convert gradient from C/cm to C/meter
  Tgrad = 100.0*Tgrad

  ! @@@@
  lscale = 10.0

  ! Write
  call get_free_unit(nout)
  open(nout,file=fout,status='unknown',err=900)
  write(nout,'(''# X Y Ts Ta Tb Toff Tgrad q_soil q_conv q_rad q_evap'')')
  do i=1,ndat
     ! Conductive heat flow
     q_soil = k_soil*Tgrad(i)
     ! Convective heat flow
     h_conv = convective_heat_coefficient(Ts(i),Ta(i),lscale)
     q_conv = h_conv*(Ts(i)-Ta(i))
     ! Convective heat flow (background)
     h_conv_b = convective_heat_coefficient(Tb(i),Ta(i),lscale)
     q_conv_b = h_conv_b*(Tb(i)-Ta(i))
     Tsky4 = Tb(i)**4
     ! Correct Tsky (convective flux)
     if(correct_tsky_conv) Tsky4 = Tsky4 - q_conv_b/(emiss_rel*sb_constant)
     ! Evaporation heat flow
     q_evap = 0.0
     if(rhum(i) <= 100.0) then
        Pair = Pair_standard      ! Atmospheric pressure
        tmean = 0.5*(Ts(i)+Ta(i)) ! Average temperature of air near the soil
        ! tmean = 303.0  ! @@@  303K = 30C
        psats = saturation_pressure_over_liquid(Ts(i)) ! Psat at soil
        psata = saturation_pressure_over_liquid(Ta(i)) ! Psat in the air
        pwata = psata*rhum(i)/100.0
        rhows = vapor_density(Ts(i),psats)
        rhowa = vapor_density(Ta(i),pwata)
        if(rhows >= rhowa) then
           lewis = air_thermal_diffusivity(tmean,Pair)/ &
                vapor_diffusion_coefficient(tmean,Pair)
           ! Use mass transfer/heat transfer analogy
           h_vapor=h_conv/(air_density(Tmean,Pair)*air_cp(Tmean)*lewis**(2./3.))
           q_evap = L_lv*h_vapor*(rhows-rhowa)
        end if
        ! Correct Tsky (evaporation flux)
        if(correct_tsky_evap) then
           tmean_b = 0.5*(Tb(i)+Ta(i)) ! Average temperature of air near bkgnd
           psatb = saturation_pressure_over_liquid(Tb(i)) ! Psat at background
           rhowb = vapor_density(Tb(i),psatb)
           if(rhowb >= rhowa) then
              lewis_b = air_thermal_diffusivity(tmean_b,Pair)/ &
                   vapor_diffusion_coefficient(tmean_b,Pair)
              h_vapor_b=h_conv_b/(air_density(Tmean_b,Pair)*air_cp(Tmean_b)* &
                   lewis_b**(2./3.))
              q_evap_b = L_lv*h_vapor_b*(rhowb-rhowa)
           end if
           Tsky4 = Tsky4 - q_evap_b/(emiss_rel*sb_constant)
        end if
     end if
     ! Radiative heat flow
     q_rad = emiss_rel*sb_constant*(Ts(i)**4-Tsky4)
     ! Eliminate outliers
     err = abs(q_soil - (q_conv+q_rad+q_evap))
     if(err > 1000.) then    ! @@@@@@@ Big error: outlier
        write(*,'(''Eliminated outlier'')')
        cycle
     end if
     write(nout,'(2(1x,f9.1),4(1x,f6.2),1x,f6.1,4(1x,f8.3))') x(i),y(i), &
          Ts(i),Ta(i),Tb(i),Toff(i),Tgrad(i),q_soil,q_conv,q_rad,q_evap
  end do
  close(nout)

  call exit(0)
900 write(*,'(''Cannot open output file: '',a)') trim(fout)
  call exit(1)
end program calcflussi
