module m_airvapor
  ! Properties of water and vapor in the atmosphere
  !
  ! Author: Giovanni Macedonio
  !
  ! Version: 1.0
  !
  ! Date: 23-SEP-2018
  !
  ! Note: some routines are extracted from module enthalphytools
  !
  implicit none
  real(8), parameter :: runiv=8.314472     ! Universal gas constant
  real(8), parameter :: pmola=0.029        ! Molecular weight of air (kg/mole)
  real(8), parameter :: pmolw=0.018        ! Molecular weight of water (kg/mole)
  real(8), parameter :: rair=runiv/pmola   ! Gas constant of air
  real(8), parameter :: rwater=runiv/pmolw ! Gas constant of water vapor
  real(8), parameter :: L_lv = 2.501e6     ! Latent heat of liquid->vapour
  real(8) :: beta_air = 3.69e-3
  real(8) :: kvisc_air = 1.35e-5
  real(8) :: grav = 9.81

contains

  function air_thermal_conductivity(temp) result(kappa)
    ! Thermal conductivity of dry air (w/mK)
    ! From Jacobson, 2005
    implicit none
    real(8), intent(in)  :: temp    ! Temperature (K)
    real(8) :: kappa      ! Thermal conductivity (W/mK)
    kappa = 0.023807 + 7.1128e-5*(temp-273.15)  ! Eq.2.5 Jacobson, 2005
  end function air_thermal_conductivity

  function vapor_thermal_conductivity(temp) result(kappa)
    ! Thermal conductivity of vapor (w/mK)
    ! From Jacobson, 2005
    implicit none
    real(8), intent(in)  :: temp    ! Temperature (K)
    real(8) :: kappa      ! Thermal conductivity (W/mK)
    kappa = 0.015606 + 8.3680e-5*(temp-273.15)  ! Eq.2.6 Jacobson, 2005
  end function vapor_thermal_conductivity

  !  function moist_air_thermal_conductivity(temp)
  !  See eq.2.7 Jacobson, 2007

  function air_thermal_diffusivity(temp,pres) result(alpha)
    ! Thermal diffusivity of dry air (m2/s)
    ! From Jacobson (2005) eq.8.14
    implicit none
    real(8), intent(in)  :: temp    ! Temperature (K)
    real(8), intent(in)  :: pres    ! Pressure (Pa)
    real(8) :: alpha      ! Thermal diffusivity (m2/s)
    alpha = air_thermal_conductivity(temp)/(air_density(temp,pres)*air_cp(temp))
  end function air_thermal_diffusivity

  function vapor_diffusion_coefficient(temp,pres) result(cdiff)
    ! Diffusion coefficient of vapor in air (m2/s)
    ! From Pruppacher and Klett (1997); Jacobson (2005)
    implicit none
    real(8), intent(in)  :: temp    ! Temperature (K)
    real(8), intent(in)  :: pres    ! Pressure (Pa)
    real(8) :: cdiff   ! Diffusion coefficient (m2/s)
    cdiff = 2.11e-5*(temp/273.15)**1.94*(1.01325e5/pres)
  end function vapor_diffusion_coefficient

  function saturation_pressure_over_liquid(T) result(psat)
    ! Vapor saturation pressure over liquid in Pa (T > Tref)
    implicit none
    real(8), intent(in)  :: T
    real(8) :: psat
    psat = 611.22*exp(17.67*(T-273.16)/(T-29.65))
  end function saturation_pressure_over_liquid

  function temperature_at_saturation_over_liquid(P) result(T)
    ! This is T(Psat): the inverse of Psat(T) for T>Tref
    ! Defined for P>0
    implicit none
    real(8), intent(in)  :: P
    real(8) :: T
    real(8) :: a
    a = 17.67/log(P/611.22)
    T = (29.65-a*273.16)/(1.0-a)
  end function temperature_at_saturation_over_liquid

  function convective_heat_coefficient(ts,ta,lscale) result(hconv)
    ! Coefficiente di scambio termico convettivo
    implicit none
    real(8), intent(in) :: ts,ta,lscale ! Tsoil, Tair, Length_scale
    real(8) :: hconv
    real(8) :: rayleigh,nusselt,tdiff_air,k_air,tmean,pres
    pres = 1.01325e5  ! Atmospheric pressure @@@@@@@@@@@@
    tmean = 0.5*(ts+ta)
    tdiff_air = air_thermal_diffusivity(tmean,pres)
    k_air = air_thermal_conductivity(tmean)
    rayleigh = grav*beta_air*abs(Ts-Ta)*lscale**3/(kvisc_air*tdiff_air)
    ! write(*,*) rayleigh
    if(Ts > Ta) then
       if(rayleigh < 1e7) then
          nusselt = 0.54*rayleigh**0.25
       else
          nusselt = 0.15*rayleigh**0.33333
       end if
    else
       nusselt = 0.27*rayleigh**0.25
    end if
    hconv = k_air*nusselt/lscale
  end function convective_heat_coefficient

  function vapor_density(temp,pres)
    ! Density of vapor (ideal gas)
    implicit none
    real(8), intent(in) :: temp,pres   ! K,Pa
    real(8) :: vapor_density           ! kg/m3
    vapor_density = pres/(rwater*temp)
  end function vapor_density

  function air_density(temp,pres)
    ! Density of air (ideal gas)
    implicit none
    real(8), intent(in) :: temp,pres   ! K,Pa
    real(8) :: air_density             ! kg/m3
    air_density = pres/(rair*temp)
  end function air_density

  function air_cp(temp)
    ! Specific heat of dry air at constant pressure (J/kgK)
    implicit none
    real(8), intent(in) :: temp   ! K
    real(8) :: air_cp
    air_cp = 1004.67  ! At 298 K
  end function air_cp

end module m_airvapor
