module parseline
  !
  ! Get fields from a string and associate variable names to variable pointers
  !
  ! Author: G.Macedonio
  !
  ! Version: 1.2
  !
  ! Date of first version: 22-NOV-2014
  ! Date of this  version: 20-JUN-2018
  !
  ! Note: write errors on unit 6
  !
  ! Provides:
  ! subroutine associate_variable(varlist,fieldname,varind,stat)
  ! function check_keyval(line,sep)
  ! subroutine get_fields(line,nfield,field)
  ! subroutine get_key(line,key,sep,stat,pos)
  ! subroutine get_keyval(line,key,sep,nvalues,values,stat)
  ! subroutine assign_fields(varlist,varmap,nfields,field_value)
  ! Generic subroutine add_to_list(varlist,string,variable,stat)
  ! subroutine add_to_list_char(varlist,string,variable,stat)
  ! subroutine add_to_list_logic(varlist,string,variable,stat)
  ! subroutine add_to_list_int4(varlist,string,variable,stat)
  ! subroutine add_to_list_real4(varlist,string,variable,stat)
  ! subroutine add_to_list_real8(varlist,string,variable,stat)
  ! subroutine allocate_list_elements(varlist,maxlen,maxvars,stat)
  ! subroutine reallocate_list_elements(varlist,newsize,stat)
  ! subroutine allomem_char(array,maxlen,n1,stat)
  ! subroutine reallomem_char(array,newsize,stat)
  ! subroutine allomem_ptr_to_variable(array,n1,stat)
  ! subroutine reallomem_ptr_to_variable(array,newsize,stat)
  ! subroutine deallomem_char(array,stat)
  ! subroutine deallomem_ptr_to_variable(array,stat)
  ! logical function isblank(char)
  ! logical function strcompare(str1,str2)
  ! subroutine uppercase(str)
  ! subroutine lowercase(str)
  !
  ! TODO: get_fields should handle strings like 'aaa nnn'
  !
  implicit none
  integer, parameter :: STRING_LENGTH=80

  integer, parameter :: TYPE_UNDEF=0
  integer, parameter :: TYPE_CHAR=1
  integer, parameter :: TYPE_INT4=2
  integer, parameter :: TYPE_REAL4=3
  integer, parameter :: TYPE_REAL8=4
  integer, parameter :: TYPE_LOGIC=5

  type :: ptr_to_variable
     integer :: ptype = TYPE_UNDEF  ! Type: char,int4,real4,real8,logical
     real(8), pointer :: real8 => NULL()
     real(4), pointer :: real4 => NULL()
     integer, pointer :: int4  => NULL()
     logical, pointer :: logic => NULL()
     ! Deferred-length character are supported in F2003
     ! character(len=:), pointer :: char
     character(len=STRING_LENGTH), pointer :: char => NULL()
  end type ptr_to_variable

  type :: list_of_variables
     integer :: maxvars = 0 ! Number of allocated elements
     integer :: nvars = 0            ! Number of elements
     ! Deferred-length character are supported in F2003
     ! character(len=:), pointer :: name(:)
     character(len=STRING_LENGTH), pointer :: name(:) => NULL()
     type(ptr_to_variable), pointer :: ptr(:) => NULL()
  end type list_of_variables

  interface add_to_list
     module procedure add_to_list_char
     module procedure add_to_list_logic
     module procedure add_to_list_int4
     module procedure add_to_list_real4
     module procedure add_to_list_real8
  end interface add_to_list

contains

  subroutine associate_variable(varlist,fieldname,varind,stat)
    ! Associate a variable to a pointer in the list
    implicit none
    type(list_of_variables) :: varlist
    character(len=*) :: fieldname
    integer :: varind
    integer :: stat
    integer :: i
    stat = 1
    do i=1,varlist%nvars
       if(strcompare(varlist%name(i),fieldname)) then
          varind = i
          stat = 0
          return
       end if
    end do
  end subroutine associate_variable

  function check_keyval(line,sep) result(ok)
    ! Check if a string contains a valid key=value(s)
    ! If sep=='' check only the key
    implicit none
    character(len=*), intent(in)  :: line
    character(len=1), intent(in)  :: sep  ! Key-value separator (usually '=')
    logical :: ok
    integer :: linelen
    integer :: i,j2
    logical :: found
    !
    ok = .false.
    linelen = len_trim(line)
    ! Find the key
    found = .false.
    do i=1,linelen
       if(isblank(line(i:i))) then
          if(.not.found) then
             cycle
          else
             exit
          end if
       else
          found = .true.
          j2 = i  ! Position of last non blank character read
          if(line(i:i) == sep) exit
       end if
    end do
    if(found) then
       ! Find separator
       if(sep == '') then
          ok = .true.
          return
       end if
       found = .false.
       do i=j2,linelen
          if(line(i:i) == sep) then
             found = .true.
             j2 = i
             exit
          end if
       end do
       if(.not.found) return  ! (ok=.false.)
       ! Find value (at least one non blank character)
       do i=j2,linelen
          if(.not.isblank(line(i:i))) then
             ok = .true.
             return
          end if
       end do
    end if
    ! Return (ok=.false.)
  end function check_keyval

  subroutine get_fields(line,nfield,field)
    ! Gets the fields from a string
    ! Memory for the fields is allocated
    implicit none
    character(len=*), intent(in)  :: line
    integer, intent(out) :: nfield
    character(len=*), allocatable :: field(:)
    integer :: linelen
    integer :: i,j
    logical :: found
    !
    linelen = len_trim(line)
    ! First pass: Count number of fields (for allocating memory)
    nfield=0
    found = .false.
    do i=1,linelen
       if(.not.isblank(line(i:i)) .and. .not.found) then
          nfield = nfield+1
          found = .true.
       end if
       if(isblank(line(i:i))) found = .false.
    end do
    if(allocated(field)) deallocate(field)
    allocate(field(nfield))
    ! Second pass: get fields
    nfield=0
    j=0
    found = .false.
    do i=1,linelen
       if(.not.isblank(line(i:i)) .and. .not.found) then
          nfield = nfield+1
          j = 1
          found = .true.
          field(nfield) = ''   ! Must clear
       end if
       if(isblank(line(i:i))) found = .false.
       if(found) then
          field(nfield)(j:j) = line(i:i)
          j = j+1
       end if
    end do
  end subroutine get_fields

  subroutine get_key(line,key,sep,stat,pos)
    ! Get the first field of a record (key)
    ! Optionally returns the position of the end of the key in line
    implicit none
    character(len=*), intent(in)  :: line
    character(len=*), intent(out) :: key
    integer, intent(out) :: stat
    character(len=1), intent(in)  :: sep  ! Key-value separator (usually '=')
    integer, optional, intent(out) :: pos ! Position of last non-blank
    integer :: linelen
    integer :: i,j1,j2
    logical :: found
    !
    linelen = len_trim(line)
    stat = 0
    ! Find the key
    found = .false.
    do i=1,linelen
       if(isblank(line(i:i))) then
          if(.not.found) then
             cycle
          else
             exit
          end if
       else
          if(.not.found) j1=i ! Position of first non blank character read
          found = .true.
          j2 = i  ! Position of last non blank character read
       end if
       if(line(i:i) == sep) exit
    end do
    if(.not.found) then
       stat = 1
       return
    end if
    key=line(j1:j2)
    if(present(pos)) pos=j2
  end subroutine get_key

  subroutine get_keyval(line,key,sep,nvalues,values,stat)
    ! Gets the key = value1 [value2, ...]
    ! If sep=='' reads only the key (same as get_key)
    implicit none
    character(len=*), intent(in)  :: line
    character(len=1), intent(in)  :: sep  ! Key-value separator (usually '=')
    character(len=*), intent(out) :: key
    integer, intent(out) :: nvalues
    character(len=*), allocatable :: values(:)
    integer, intent(out) :: stat
    integer :: linelen,keylen
    integer :: i,j1,j2
    logical :: found
    !
    linelen = len_trim(line)
    stat = 0
    call get_key(line,key,sep,stat,j2)
    if(stat /= 0) return
    ! Find separator
    if(sep == '') then
       nvalues = 0
       return
    end if
    keylen = len_trim(key)
    if(key(keylen:keylen) == sep) then
       ! Key contains the separator
       key(keylen:keylen) = ' '
       j2 = max(0,j2-1)
    end if
    do i=j2+1,linelen
       found = .false.
       j1 = i
       if(line(i:i) == sep) then
          found = .true.
          exit
       end if
    end do
    if(.not.found) then
       stat = 2
       return
    end if
    ! Get values
    call get_fields(line(j1+1:linelen),nvalues,values)
  end subroutine get_keyval

  subroutine assign_fields(varlist,varmap,nfields,field)
    ! Assegna i valori dei campi alle variabili secondo la varmap
    implicit none
    integer, intent(in) :: nfields
    integer :: varmap(nfields)
    character(len=*), intent(in) :: field(:)
    type(list_of_variables), intent(inout) :: varlist
    integer :: i
    if(.not.associated(varlist%name)) return
    do i=1,nfields
       ! if(varmap(i) > varlist%nvars) call exit(1)
       select case(varlist%ptr(varmap(i))%ptype)
       case(TYPE_CHAR)
          varlist%ptr(varmap(i))%char =  field(i)
       case(TYPE_INT4)
          read(field(i),*) varlist%ptr(varmap(i))%int4
       case(TYPE_REAL4)
          read(field(i),*) varlist%ptr(varmap(i))%real4
       case(TYPE_REAL8)
          read(field(i),*) varlist%ptr(varmap(i))%real8
       case(TYPE_LOGIC)
          read(field(i),*) varlist%ptr(varmap(i))%logic
       end select
    end do
  end subroutine assign_fields

  subroutine add_to_list_char(varlist,string,variable,stat)
    implicit none
    integer, parameter :: N_ALLOCATE=8
    type(list_of_variables), intent(inout) :: varlist
    character(len=*), intent(in) :: string
    character(len=*), target :: variable
    integer, intent(out), optional :: stat
    integer :: maxlen = STRING_LENGTH
    integer :: istat
    if(present(stat)) stat = 0
    !
    if(varlist%maxvars == 0) then  ! First time memory allocation
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call allocate_list_elements(varlist,maxlen,varlist%maxvars,istat)
       if(istat /= 0 .and. present(stat)) stat = stat+1
    end if
    if(varlist%nvars >= varlist%maxvars) then
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call reallocate_list_elements(varlist,varlist%maxvars,istat)
       if(istat /= 0 .and. present(stat)) stat = stat+2
    end if
    varlist%nvars = varlist%nvars + 1
    varlist%name(varlist%nvars)  = ''
    varlist%name(varlist%nvars)  = string
    varlist%ptr(varlist%nvars)%char => variable
    varlist%ptr(varlist%nvars)%ptype = TYPE_CHAR
  end subroutine add_to_list_char

  subroutine add_to_list_logic(varlist,string,variable,stat)
    implicit none
    integer, parameter :: N_ALLOCATE=8
    type(list_of_variables), intent(inout) :: varlist
    character(len=*), intent(in) :: string
    logical, intent(in), target  :: variable
    integer, intent(out), optional :: stat
    integer :: maxlen = STRING_LENGTH
    integer :: istat
    !
    if(varlist%maxvars == 0) then  ! First time memory allocation
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call allocate_list_elements(varlist,maxlen,varlist%maxvars,stat)
    end if
    if(varlist%nvars >= varlist%maxvars) then
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call reallocate_list_elements(varlist,varlist%maxvars,istat)
       if(present(stat)) stat = istat
    end if
    varlist%nvars = varlist%nvars + 1
    varlist%name(varlist%nvars)  = ''
    varlist%name(varlist%nvars)  = string
    varlist%ptr(varlist%nvars)%logic => variable
    varlist%ptr(varlist%nvars)%ptype = TYPE_LOGIC
  end subroutine add_to_list_logic

  subroutine add_to_list_int4(varlist,string,variable,stat)
    implicit none
    integer, parameter :: N_ALLOCATE=8
    type(list_of_variables) :: varlist
    character(len=*) :: string
    integer, target  :: variable
    integer, intent(out), optional :: stat
    integer :: maxlen = STRING_LENGTH
    integer :: istat
    !
    if(varlist%maxvars == 0) then  ! First time memory allocation
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call allocate_list_elements(varlist,maxlen,varlist%maxvars,istat)
       if(istat /= 0) then
          if(present(stat)) stat = 1
          return
       end if
    end if
    if(varlist%nvars >= varlist%maxvars) then
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call reallocate_list_elements(varlist,varlist%maxvars,istat)
       if(istat /= 0) then
          if(present(stat)) stat = 2
          return
       end if
    end if
    varlist%nvars = varlist%nvars + 1
    varlist%name(varlist%nvars)  = ''
    varlist%name(varlist%nvars)  = string
    varlist%ptr(varlist%nvars)%int4 => variable
    varlist%ptr(varlist%nvars)%ptype = TYPE_INT4
  end subroutine add_to_list_int4

  subroutine add_to_list_real4(varlist,string,variable,stat)
    implicit none
    integer, parameter :: N_ALLOCATE=8
    type(list_of_variables) :: varlist
    character(len=*) :: string
    real, target    :: variable
    integer, intent(out), optional :: stat
    integer :: maxlen = STRING_LENGTH
    integer :: istat
    !
    if(varlist%maxvars == 0) then  ! First time memory allocation
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call allocate_list_elements(varlist,maxlen,varlist%maxvars,istat)
       if(istat /= 0) then
          if(present(stat)) stat = 1
          return
       end if
    end if
    if(varlist%nvars >= varlist%maxvars) then
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call reallocate_list_elements(varlist,varlist%maxvars,istat)
       if(istat /= 0) then
          if(present(stat)) stat = 2
          return
       end if
    end if
    varlist%nvars = varlist%nvars + 1
    varlist%name(varlist%nvars)  = ''
    varlist%name(varlist%nvars)  = string
    varlist%ptr(varlist%nvars)%real4 => variable
    varlist%ptr(varlist%nvars)%ptype = TYPE_REAL4
  end subroutine add_to_list_real4

  subroutine add_to_list_real8(varlist,string,variable,stat)
    implicit none
    integer, parameter :: N_ALLOCATE=8
    type(list_of_variables) :: varlist
    character(len=*) :: string
    real(8), target    :: variable
    integer, intent(out), optional :: stat
    integer :: maxlen = STRING_LENGTH
    integer :: istat
    !
    if(varlist%maxvars == 0) then  ! First time memory allocation
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call allocate_list_elements(varlist,maxlen,varlist%maxvars,istat)
       if(istat /= 0) then
          if(present(stat)) stat = 1
          return
       end if
    end if
    if(varlist%nvars >= varlist%maxvars) then
       varlist%maxvars = varlist%maxvars + N_ALLOCATE
       call reallocate_list_elements(varlist,varlist%maxvars,istat)
       if(istat /= 0) then
          if(present(stat)) stat = 2
          return
       end if
    end if
    varlist%nvars = varlist%nvars + 1
    varlist%name(varlist%nvars)  = ''
    varlist%name(varlist%nvars)  = string
    varlist%ptr(varlist%nvars)%real8 => variable
    varlist%ptr(varlist%nvars)%ptype = TYPE_REAL8
  end subroutine add_to_list_real8

  subroutine allocate_list_elements(varlist,maxlen,maxvars,stat)
    ! Allocates varlist%maxvars elements
    implicit none
    type(list_of_variables) :: varlist
    integer, intent(in) :: maxlen      ! Maximum name length
    integer, intent(in) :: maxvars     ! Maximum number of variables
    integer, intent(out) :: stat
    call allomem_char(varlist%name,maxlen,maxvars,stat)
    call allomem_ptr_to_variable(varlist%ptr,maxvars,stat)
    varlist%maxvars = maxvars
  end subroutine allocate_list_elements

  subroutine reallocate_list_elements(varlist,newsize,stat)
    ! Reallocates varlist%maxvars elements
    implicit none
    type(list_of_variables) :: varlist
    integer, intent(in)  :: newsize
    integer, intent(out) :: stat
    call reallomem_ptr_to_variable(varlist%ptr,newsize,stat)
    call reallomem_char(varlist%name,newsize,stat)
  end subroutine reallocate_list_elements

  !
  ! Memory allocation/reallocation/deallocation
  !
  subroutine allomem_char(array,maxlen,nelem,stat)
    ! Allocate memory
    implicit none
    integer, intent(in) :: maxlen
    character(len=maxlen), pointer :: array(:)
    integer, intent(in) :: nelem
    integer, intent(out) :: stat
    allocate(array(nelem),STAT=stat)
  end subroutine allomem_char

  subroutine reallomem_char(array,newsize,stat)
    ! Reallocate memory
    implicit none
    character(len=*), pointer :: array(:)
    integer, intent(in) :: newsize
    character(len=len(array)), pointer :: p(:)
    integer :: oldsize,minsize
    integer, intent(out) :: stat
    integer :: chlen
    chlen = len(array)
    oldsize = size(array)
    minsize = min(oldsize,newsize)
    allocate(p(newsize),STAT=stat)
    ! Copy
    p(1:minsize) = array(1:minsize)
    ! Clear remaining memory
    if(minsize+1 < newsize) p(minsize+1:newsize) = ' '
    deallocate(array)
    array => p
  end subroutine reallomem_char

  subroutine allomem_ptr_to_variable(array,n1,stat)
    ! Allocate memory
    implicit none
    type(ptr_to_variable), pointer :: array(:)
    integer, intent(in) :: n1
    integer, intent(out) :: stat
    integer, parameter :: element_size=4
    allocate(array(n1),STAT=stat)
  end subroutine allomem_ptr_to_variable

  subroutine reallomem_ptr_to_variable(array,newsize,stat)
    ! Reallocate memory
    implicit none
    type(ptr_to_variable), pointer :: array(:)
    type(ptr_to_variable), pointer :: p(:)
    integer, intent(in) :: newsize
    integer :: oldsize,minsize
    integer, intent(out) :: stat
    integer, parameter :: element_size=4
    oldsize = size(array)
    minsize = min(oldsize,newsize)
    allocate(p(newsize),STAT=stat)
    p(1:minsize) = array(1:minsize)  ! Copy
    deallocate(array)
    array => p
  end subroutine reallomem_ptr_to_variable

  subroutine deallomem_char(array,stat)
    ! Deallocate memory
    implicit none
    character(len=*), pointer :: array(:)
    integer, intent(out) :: stat
    integer :: chlen
    if(associated(array)) then
       chlen = len(array)
       deallocate(array,STAT=stat)
    endif
  end subroutine deallomem_char

  subroutine deallomem_ptr_to_variable(array,stat)
    ! Deallocate memory
    implicit none
    type(ptr_to_variable), pointer :: array(:)
    integer, intent(out) :: stat
    if(associated(array)) then
       deallocate(array,STAT=stat)
    endif
  end subroutine deallomem_ptr_to_variable

  logical function isblank(char)
    ! Check if a character is space or tab
    implicit none
    character, intent(in) :: char
    isblank = .false.
    if(ichar(char) == 9 .or. ichar(char) == 32) isblank = .true.
  end function isblank

  logical function strcompare(str1,str2)
    ! Compare two strings
    ! Return .true. if the strings are equal
    !
    ! WARNING: Strings are first converted to uppercase
    !
    character(len=*), intent(in) :: str1,str2
    character(len=len(str1)) :: str1_copy
    character(len=len(str2)) :: str2_copy
    !
    str1_copy = str1
    str2_copy = str2
    call uppercase(str1_copy)
    call uppercase(str2_copy)
    !
    if(TRIM(str1_copy) == TRIM(str2_copy)) then
       strcompare = .true.
    else
       strcompare = .false.
    endif
  end function strcompare

  subroutine uppercase(str)
    ! Convert a string to uppercase
    implicit none
    character(len=*), intent(inout) :: str
    integer :: i, del
    !
    del = iachar('a') - iachar('A')
    do i = 1, len_trim(str)
       if (lge(str(i:i),'a') .and. lle(str(i:i),'z')) then
          str(i:i) = achar(iachar(str(i:i)) - del)
       end if
    end do
  end subroutine uppercase

  subroutine lowercase(str)
    ! Convert a string to lowercase
    implicit none
    character(len=*), intent(inout) :: str
    integer :: i, del
    !
    del = iachar('a') - iachar('A')
    do i = 1, len_trim(str)
       if (lge(str(i:i),'A') .and. lle(str(i:i),'Z')) then
          str(i:i) = achar(iachar(str(i:i)) + del)
       end if
    end do
  end subroutine lowercase

end module parseline
