module timeutil
  !
  ! Operations with date and time
  !
  ! Author: G.Macedonio
  !
  ! Version: 1.1
  !
  ! Date of first version:  1-AUG-2012
  ! Date of this  version: 22-JUN-2018
  !
  ! Note: These routines do not consider the leap seconds.
  ! In fact, a day is not -exactly- 86400 seconds long, and
  ! periodically seconds are added to days in order to keep time
  ! synchronized with the various observable standards.
  ! These are called Leap Seconds (https://en.wikipedia.org/wiki/Leap_second)
  ! Between 1970 and 2013 25 leap seconds were added
  !
  ! ATTENZIONE:
  ! La Julian date (JD) inizia alle ore 12:00 del giorno Gregoriano
  ! Il Julian day number (JDN) e' il numero intero associato al JD
  !
  ! In queste routines si considera il JDN che inizia alle ore 12:00 del
  ! giorno greogoriano corrispondente.
  !
  ! Esempio: il 31-12-1999 alle ore 12:00 inizia il JD 2451544.0
  ! Alle ore 24:00 del del 31-12-1999, corrispondenti alle 00:00 del 1-1-2000,
  ! si ha JD=2451544.5, il JDN e' 2451544
  ! Il MJD inizia alle ore 00:00, dove MJD=JD-2400000.5
  ! Il giorno 1-1-2000 ore 00:00 inizia il MJD 51544
  !
  ! This module provides:
  ! real(8) function roundoff(number,ndigit)
  ! subroutine jdn2gregorian(jdn,year,month,day)
  ! subroutine gregorian2jdn(jdn,year,month,day)
  ! subroutine jdate2gregorian(jdate,year,month,day,hour,minute,seconds)
  ! subroutine gregorian2jdate(jdate,year,month,day,hour,minute,seconds)
  ! subroutine gregorian2mjd(mjd,year,month,day,hour,minute,seconds)
  ! subroutine mjd2gregorian(mjd,year,month,day,hour,minute,seconds)
  ! subroutine days_per_month(year,month,dpm)
  ! subroutine encode_time(enctime,year,month,day,hour,minute,second)
  ! subroutine decode_time(enctime,year,month,day,hour,minute,second)
  ! subroutine epoch_to_enctime(epoch,enctime)
  ! subroutine enctime_to_epoch(epoch,enctime)
  ! subroutine gregorian2epoch(epoch,year,month,day,hour,minute,second)
  ! subroutine epoch2gregorian(epoch,year,month,day,hour,minute,second)
  ! subroutine day2mjdn(mjdn,day,month,year)
  ! subroutine mjdn2day(mjdn,day,month,year)
  ! subroutine fracday2hms(fracday,hour,minute,seconds)
  ! subroutine hms2fracday(fracday,hour,minute,seconds)
  ! subroutine timestring(string,hour,minute,seconds,ndigit)
  !
  implicit none

contains

  real(8) function roundoff(number,ndigit)
    ! Round the number to ndigit decimal digits
    implicit none
    real(8), intent(in) :: number
    integer, intent(in) :: ndigit
    real(8) :: rs
    rs = 10d0**ndigit
    roundoff = anint(number*rs)/rs
  end function roundoff

  subroutine jdn2gregorian(jdn,year,month,day)
    !
    !  Trasforma il giorno Juliano (Julian Day Number) in giorno,mese,anno.
    !
    !  Nota: Il Lunedi 1 Gennaio 1753 e' il giorno Juliano numero 2361331.
    !
    ! Il giorno giuliano inizia alle ore 12:00 del 1.1 del 4713 (BC) del
    ! calendario Juliano prolettico, ovvero il 24.11 del 4714 (BC) del
    ! caledario Gregoriano prolettico. Attenzione: l'anno zero non esiste.
    ! Il calendario prolettico e' l'estensione del calendario indietro nel
    ! tempo, prima della sua invenzione.
    !
    implicit none
    integer,intent(in)  :: jdn          ! Julian Day Number (input)
    integer,intent(out) :: year         ! Anno (output)
    integer,intent(out) :: month        ! Mese (output)
    integer,intent(out) :: day          ! Giorno del mese (output)
    integer :: j,g,dg,c,dc,b,db,a,da,y,m,d
    j  = jdn + 32044       ! Shift back to 4800 BC
    g  = j/146097
    dg = mod(j,146097)
    c  = (dg/36524 + 1)*3/4
    dc = dg - c*36524
    b = dc/1461
    db = mod(dc,1461)
    a  = (db/365+1)*3/4
    da = db-a*365
    y = g*400+c*100+b*4+a
    m = (da*5+308)/153-2
    d = da-(m+4)*153/5+122
    year = y - 4800+(m+2)/12
    month = mod(m+2,12)+1
    day = d+1
  end subroutine jdn2gregorian

  subroutine gregorian2jdn(jdn,year,month,day)
    !
    !  Trasforma il giorno,mese,anno in Julian Day Number
    !
    !  Nota: Il Lunedi 1 Gennaio 1753 e' il giorno Juliano numero 2361331.
    !
    implicit none
    integer, intent(out) :: jdn        ! Julian Day Number (output)
    integer,intent(in)   :: year       ! Year (input)
    integer,intent(in)   :: month      ! Month (input)
    integer,intent(in)   :: day        ! Day of month (input)
    integer :: a,y,m
    a = (14-month)/12
    y = year+4800-a
    m = month+12*a-3
    jdn = day + (153*m + 2)/5 + 365*y + y/4-y/100+y/400-32045
  end subroutine gregorian2jdn

  subroutine jdate2gregorian(jdate,year,month,day,hour,minute,seconds)
    ! Convert Julian date to gregorian date
    ! Julian day begins at 12:00 Example: 2000-01-01 12:00 is JD=2451545.0
    implicit none
    real(8), intent(in)  :: jdate
    integer, intent(out) :: year,month,day,hour,minute
    real(8), intent(out) :: seconds
    integer :: j,g,dg,c,dc,b,db,a,da,y,m,d
    real(8) :: rjul,rhour,rminute
    rjul = jdate+0.5d0
    j  = int(rjul) + 32044   ! Shift back to 4800 BC
    g  = j/146097
    dg = mod(j,146097)
    c  = (dg/36524 + 1)*3/4
    dc = dg - c*36524
    b  = dc/1461
    db = mod(dc,1461)
    a  = (db/365+1)*3/4
    da = db-a*365
    y = g*400+c*100+b*4+a
    m = (da*5+308)/153-2
    d = da-(m+4)*153/5+122
    year = y - 4800+(m+2)/12
    month = mod(m+2,12)+1
    day = d+1
    ! Add time
    rhour=(rjul-aint(rjul))*24d0
    hour = int(rhour)
    rminute = (rhour-dble(hour))*60d0
    minute  = int(roundoff(rminute,5))
    seconds = max((rminute-dble(minute))*60d0,0d0)
  end subroutine jdate2gregorian

  subroutine gregorian2jdate(jdate,year,month,day,hour,minute,seconds)
    ! Convert Gregorian to Julian date
    ! Julian day begins at 12:00 Example: 2000-01-01 12:00 is JD=2451545.0
    implicit none
    real(8), intent(out) :: jdate
    integer, intent(in)  :: year,month,day,hour,minute
    real(8), intent(in)  :: seconds
    integer :: a,y,m,ijd
    logical :: gregorian=.true.  ! TODO: automatic or optional parameter
    a = (14-month)/12
    y = year+4800-a
    m = month+12*a-3
    ijd = day + (153*m + 2)/5 + 365*y + y/4
    if(gregorian) then 
       ijd = ijd-y/100+y/400-32045  ! Gregorian period
    else
       ijd = ijd-32083              ! Julian period
    end if
    jdate = dble(ijd)+dble(hour-12)/24d0+dble(minute)/1440d0+seconds/86400d0
  end subroutine gregorian2jdate

  subroutine gregorian2mjd(mjd,year,month,day,hour,minute,seconds)
    ! Convert Gregorian to Modified Julian Date
    ! Consider only for the Gregorian period
    ! 2000-01-01 is MJD 51544
    implicit none
    real(8), intent(out) :: mjd
    integer, intent(in)  :: year,month,day,hour,minute
    real(8), intent(in)  :: seconds
    integer :: a,y,m,ijd
    a = (14-month)/12
    y = year-1600-a
    m = month+12*a-3
    ijd = day + (153*m + 2)/5 + 365*y + y/4 -y/100 + y/400 - 94494
    mjd = dble(ijd)+dble(hour)/24d0+dble(minute)/1440d0+seconds/86400d0
  end subroutine gregorian2mjd

  subroutine mjd2gregorian(mjd,year,month,day,hour,minute,seconds)
    ! Convert MJD to gregorian date /Time
    implicit none
    real(8), intent(in)  :: mjd
    integer, intent(out) :: year,month,day,hour,minute
    real(8), intent(out) :: seconds
    real(8) :: jdate
    jdate = mjd+2400000.5d0
    call jdate2gregorian(jdate,year,month,day,hour,minute,seconds)
  end subroutine mjd2gregorian

  subroutine days_per_month(year,month,dpm)
    ! Return the number of days in the month
    implicit none
    integer, intent(in)  :: month,year
    integer, intent(out) :: dpm
    integer, parameter   :: ngm(12)=(/31,28,31,30,31,30,31,31,30,31,30,31/)
    dpm = ngm(month)
    if((mod(year,100) /= 0 .or. year/400==0) .and. year/4==0 .and. month==2) dpm = dpm +1
  end subroutine days_per_month

  subroutine encode_time(enctime,year,month,day,hour,minute,second)
    ! Encode time (1 second steps)
    implicit none
    integer(8), intent(out) :: enctime
    integer, intent(in)     :: year,month,day,hour,minute,second
    integer :: jdn
    call gregorian2jdn(jdn,year,month,day)
    enctime = 60_8*(60_8*(24_8*jdn + hour) + minute) + second
  end subroutine encode_time

  subroutine decode_time(enctime,year,month,day,hour,minute,second)
    ! Decode time
    implicit none
    integer(8), intent(in) :: enctime
    integer, intent(out)   :: year,month,day,hour,minute,second
    integer :: jdn
    jdn = int(enctime/86400_8)
    call jdn2gregorian(jdn,year,month,day)
    hour = int(enctime/3600_8 - jdn*24d0)
    minute = int(enctime/60_8 - jdn*1440_8 - hour*60_8)
    second = int(enctime-jdn*86400_8 - hour*3600_8 - minute*60_8)
  end subroutine decode_time

  subroutine epoch_to_enctime(epoch,enctime)
    ! Convert the unix epoch (seconds since 1970-1-1) to encoded time
    implicit none 
    integer(8), intent(in)  :: epoch
    integer(8), intent(out) :: enctime
    enctime = epoch + 210866803200_8
  end subroutine epoch_to_enctime

  subroutine enctime_to_epoch(epoch,enctime)
    ! Convert the encoded time to unix epoch (seconds since 1970-1-1)
    implicit none 
    integer(8), intent(out) :: epoch
    integer(8), intent(in)  :: enctime
    epoch = enctime - 210866803200_8
  end subroutine enctime_to_epoch

  subroutine gregorian2epoch(epoch,year,month,day,hour,minute,second)
    ! Convert date to epoch (unix time)
    implicit none
    integer(8), intent(out) :: epoch
    integer, intent(in)     :: year,month,day,hour,minute,second
    integer(8) :: enctime
    call encode_time(enctime,year,month,day,hour,minute,second)
    call enctime_to_epoch(epoch,enctime)
  end subroutine gregorian2epoch

  subroutine epoch2gregorian(epoch,year,month,day,hour,minute,second)
    ! Convert epoch (unix time) to date
    implicit none
    integer(8), intent(in) :: epoch
    integer, intent(out)   :: year,month,day,hour,minute,second
    integer(8) :: enctime
    call epoch_to_enctime(epoch,enctime)
    call decode_time(enctime,year,month,day,hour,minute,second)
  end subroutine epoch2gregorian

  subroutine day2mjdn(mjdn,day,month,year)
    ! Convert date to MJD Number (Modified Julian day number)
    ! MJD is similar to the julian day (JD), but begins at 00:00 instead
    ! of 12:00. Starting from 17.11.1858, it coincides with the julian day
    ! without the starting digits '24':
    ! MJD = JD - 2400000.5
    implicit none
    integer, intent(out) :: mjdn       ! MJDN (output)
    integer,intent(in)   :: day        ! Day of month (input)
    integer,intent(in)   :: month      ! Month (input)
    integer,intent(in)   :: year       ! Year (input)
    integer :: jdn
    call gregorian2jdn(jdn,year,month,day)
    mjdn = jdn-2400000-1
  end subroutine day2mjdn

  subroutine mjdn2day(mjdn,day,month,year)
    ! Convert MJD Number to day, month, year
    implicit none
    integer,intent(in)  :: mjdn         ! Modified Julian Day (input)
    integer,intent(out) :: day          ! Giorno del mese (output)
    integer,intent(out) :: month        ! Mese (output)
    integer,intent(out) :: year         ! Anno (output)
    integer :: jdn
    jdn = mjdn+2400000+1
    call jdn2gregorian(jdn,year,month,day)
  end subroutine mjdn2day

  subroutine fracday2hms(fracday,hour,minute,seconds)
    ! Convert fraction of day to hour, minute, seconds
    implicit none
    real(8), intent(in)  :: fracday
    integer, intent(out) :: hour
    integer, intent(out) :: minute
    real(8), intent(out) :: seconds
    seconds = fracday*86400d0
    hour = int(seconds/24d0)
    seconds = seconds - hour*24d0
    minute  = int(seconds/60d0)
    seconds = seconds - minute*60d0
  end subroutine fracday2hms

  subroutine hms2fracday(fracday,hour,minute,seconds)
    ! Convert hour, minute, seconds to fraction of day
    implicit none
    real(8), intent(out) :: fracday
    integer, intent(in)  :: hour
    integer, intent(in)  :: minute
    real(8), intent(in)  :: seconds
    fracday = (3600d0*dble(hour)+60d0*dble(minute)+seconds)/86400d0
  end subroutine hms2fracday

  subroutine timestring(string,year,month,day,hour,minute,seconds,ndigit)
    ! Write time in a string in the form YYYY MM DD HH:MM:SS.sss where the
    ! number of decimals is ndigit.
    ! The string length must be at least:  20+ndigit
    implicit none
    character(len=*), intent(out) :: string
    integer, intent(in) :: year,month,day,hour,minute,ndigit
    real(8), intent(in) :: seconds
    integer :: tyear,tmonth,tday,thour,tminute
    real(8) :: jdate,tseconds
    character(len=48) :: form
    integer :: nd
    nd = min(max(ndigit,0),6)
    tseconds = roundoff(seconds,nd)
    call gregorian2jdate(jdate,year,month,day,hour,minute,tseconds)
    call jdate2gregorian(jdate,tyear,tmonth,tday,thour,tminute,tseconds)
    if(ndigit<=0) then
       write(form,'(''(i4.4,1x,i2.2,1x,i2.2,1x,i2.2,":",i2.2,":",i2.2)'')')
       write(string,form) tyear,tmonth,tday,thour,tminute,nint(tseconds)
    else
       write(form,'(''(i4.4,1x,i2.2,1x,i2.2,1x,i2.2,":",i2.2,":",f'',i1,''.'',i1,'')'')') nd+3,nd
       write(string,form) tyear,tmonth,tday,thour,tminute,tseconds
    end if
    if(string(18:18) == ' ') string(18:18) = '0'
  end subroutine timestring

end module timeutil
